// 150709 points on the sphere, farthest from each other as far as possible
// (c) Sergey Khashin http://math.ivanovo.ac.ru/dalgebra/Khashin/index.html
//     Gleb Shkryabin
//
#pragma warning(disable : 4996)

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>

//============================================================================
void  vdbl_div(int N, double *a, double k)				// A /= k ----------------------
{
	for( int i=N; i>0; i-- ) (*a++) /= k;
}
double vdbl_len  (int N, double *a)			// |A|
{
	double res=0;
	for( int i=N; i>0; i-- ) 
	{
		double d1 = (*a++);
		res += d1*d1;
	}
	return sqrt(res);
}
double vdbl_dist(int N, double *a, double *b)			// |A-B| ----------------------
{
	double res=0;
	for( int i=N; i>0; i-- ) 
	{
		double d1 = (*a++) - (*b++);
		res += d1*d1;
	}
	return sqrt(res);
}
//============================================================================
double *A(int k);			// pointer to the k-th vector
void setRandom();			// fill array arr by random vectors
int  saveArr(char *fname);	// save data to fname
double minDist( int &i1, int &i2);	// minimal distance beetween points. i1,i2 - their indexes
double stepN( int N, double eps );	// N steps with given eps
//----------------------------------------------------------------------------
int dim=0;				// dimension of the space
int Np =0;				// number of points 
double *arr;			// array of vectors
int cnt=0;
//----------------------------------------------------------------------------
double *A(int k)
{
	return arr+k*dim;
}
//----------------------------------------------------------------------------
void setRandom() // fill array arr by random vectors
{
	srand( (unsigned)time( NULL ) );
	//srand( 777 );
	for( int i=0; i<Np; i++ ) {
		double *v = A(i);
		for( int j=0; j<dim; j++ ) {
			v[j] = (2.0*rand())/(RAND_MAX + 1)-1;
		}
		vdbl_div( dim, v, vdbl_len(dim, v));
	}
} // void setRandom() // fill array arr by random vectors
//----------------------------------------------------------------------------
int saveArr( char *fname) // save data to fname
{
	FILE *f = fopen(fname, "w");
	if( !f ) return 1;
	for( int i=0; i<Np; i++ ) {
		double *v = A(i);
		for( int j=0; j<dim; j++ ) {
			fprintf(f, " %18.15lf", v[j] );
		}
		fprintf(f, "\n" );
	}
	fclose(f);
	return 0;
}//saveArr( char *fname) // save data to fname
//----------------------------------------------------------------------------
double minDist( int &i1, int &i2) // minimal distance beetween points. i1,i2 - their indexes
{
	i1=i2=-1;
	double dMin=10, d;
	for( int j1=0; j1<Np-1; j1++ )
	for( int j2=j1+1; j2<Np; j2++ ) {
		d = vdbl_dist(dim, A(j1), A(j2) );
		if( d<dMin ) { dMin=d; i1=j1; i2=j2; }
	}
	return dMin;
}//double minDist( int &i1, int &i2) // minimal distance beetween points. i1,i2 - their indexes
//----------------------------------------------------------------------------
double stepN( int N, double eps ) // N steps with given eps
{
	int i1, i2;
	double d;
	for( int i=0; i<N; i++ ) {
		d = minDist(i1,i2);
		double *v1 = A(i1);
		double *v2 = A(i2);
		for( int j=0; j<dim; j++ ) {
			double x1=v1[j], x2 = v2[j];
			v1[j] = (1+eps)*x1 - eps*x2;
			v2[j] = (1+eps)*x2 - eps*x1;
		}
		vdbl_div( dim, v1, vdbl_len(dim, v1));
		vdbl_div( dim, v2, vdbl_len(dim, v2));
	}
	cnt += N;
	return d;
}//double stepN( int N, double eps ) // N steps with given eps
//----------------------------------------------------------------------------
static void test_01_help(char* msg)
{
	printf("%%3 := list of optimal %%2 points on sphere in %%1-dimensional space\n");
	printf("\n%s\n", msg);
	_exit(0);
}
//----------------------------------------------------------------------------
int main(int argc, char* argv[])
{
	if( argc != 4 ) test_01_help("Illegal argc");
	dim = atoi(argv[1]);		// space dimension
	if( dim<1 || dim>10000000 ) test_01_help("Illegal dim(%%1)");
	Np = atoi(argv[2]);			// number of points 
	if( Np<1 || Np>10000000 ) test_01_help("Illegal number of points(%%2)");

	printf("dim=%d, N points=%d ", dim, Np); 
	arr = new double[dim*Np];
	setRandom();

	double eps=0.1, d0, d=2.01;
	while( eps>1e-12 ) {	
		do {
			d0 = d;
			d = stepN(10*dim, eps);
		} while( d>d0+eps);
		//printf("cnt=%9d, d=%18.15lf, eps=%10.2e\n", cnt, d, eps); 
		eps *= 0.57;
	}

	printf("minDist=%18.15lf\n", d); 
	if( saveArr(argv[3])!= 0) test_01_help("Illegal output file (%%3)");;
	delete []arr;
	return 0;
}

